// SeisAD18.cpp

#include "stdafx.h"
#include <aygshell.h>
//#define STRICT
#include <windows.h>
#include <stdio.h>
#include "seislog.h"
#include "library.h"
#include "msglist.h"
#include "resource.h"
#include "process.h"
#include "eventlog.h"

#define WINCE
#define MAX_SMPLRATE 200
#define MAX_DIGITIZERS 4
#define MAX_HEADERBYTES 18

//
// Definitions related to the SeisAD18Buffer class.
//

struct DIG_HEADER {
   BYTE Header[3];                  // First 3 bytes of header.
   BYTE DigNr;                      // Digitizer number.
   BYTE ChkSum1;                    // Checksum for first channel.
   BYTE ChkSum2;                    // Checksum for second channel.
   BYTE ChkSum3;                    // Checksum for third channel.
   BYTE ChkSum4;                    // Always zero.
   BYTE SmapleRate;                 // Sample rate.
   BYTE Gain;                       // Gain.
   BYTE GPSStatus;                  // 0=1PPS present, 1=1PPS not present.
};


struct CHNLDATA {                   // This structure holds channel spesific information.
   WORD NumBytes;                   // Number of data-bytes currently in this buffer.
   BYTE Bytes[MAX_SMPLRATE*3];      // Data bytes for this channel.   
};

struct DIGHEADER {                  // This structure holds digitizer header information.
   BYTE ChkSum1;                    // Calculated checksum for this digitizer (byte1).
   BYTE ChkSum2;                    // Calculated checksum for this digitizer (byte2).
   BYTE ChkSum3;                    // Calculated checksum for this digitizer (byte3).
   WORD NumBytes;                   // Number of header-bytes currently in this buffer.
   BYTE Bytes[MAX_HEADERBYTES];     // Header bytes for this digitizer (room for 18 bytes).   
};


class SeisAD18Buffer {
private:
   WORD SmplRate;                         // The samplerate used by the device.
   BYTE NumChnls;                         // Number of channels on device.
   BYTE NumDig;                           // Number of digitizers on device.
   BYTE CurDigNr;                         // A digitizer number.
   BYTE ChnlNr;                           // A channel number.
   BYTE CurSmplNr;                        // A sample number.
   WORD Index;                            // Used by several member functions.
   DIGHEADER HeaderBytes[MAX_DIGITIZERS]; // Header bytes from each digitizer.
   CHNLDATA ChnlData[MAX_DIGITIZERS*3];   // Channel information (Samples).


public:
   VOID InitBuffer(BYTE, WORD);	               // Initialize the buffer.
   BYTE Insert12BytesBlock(PCHAR);              // Insert 12 bytes block into buffer.
   VOID ResetBuffer(VOID);			               // Clears the buffer contents.
   VOID InsertBuffer(PBYTE);                    // Insert 12 byte block into buffer. 
   BOOL IsBufferFull(VOID);                     // Return TRUE if buffer holds 1 sec of data.
   BOOL VerifyChkSum(BYTE, BYTE, BYTE, BYTE);   // Compare given chksums against header chksums.
   VOID GetChkSum(BYTE, PBYTE, PBYTE, PBYTE);   // Returns the header checksums.
   VOID CopyChnlData(BYTE, PBYTE);              // Copies data for one channel to a buffer.
   BYTE GetGPSStatus(BYTE);                     // Returns the GPS status for a digitizer.
};




//
// This structure is read from the binary device config file.
//
struct DEVICE_CONFIG {
   WORD DeviceID;             // Device-ID for this device is 210.
   BYTE NumChannels;          // Can be 3,6,9 or 12.
   TCHAR PortName[6];          // Serial port COM1->COM16.
   BYTE ConfigNr;					// Spesifies the operational configuration (1-24).
   DOUBLE TimeCorrection;     // Time correction in seconds.
};


//
// This structure is used to pass startup- 
// information to the serial-reader thread.
//
struct STARTINFO_DEVREADER { 
   HANDLE hComm;              // Handle to serial port. 
   TCHAR CommSpeed[7];		   // Communication speed.
   DWORD ThreadID;            // Thread-ID for parent thread.
   BOOL *SerialRead;          // The serial-reader thread will exit if this becomes FALSE.
   DEVMNGR_INTERF_STRUCT *DS; // Pointer to the exported structure.
   DWORD* BlockNr;            // Pointer to variable who are to receive block numbers.
   DOUBLE TimeCorrection;     // Time correction to apply (in seconds).
   BYTE ConfigNr;					// Configuration number.
   BYTE SyncStatus;           // The current status of the time synchronization thread.
   HANDLE hSyncStatusMutex;   // Handle to a mutex object used to control access to the
                              // SyncStatus variable in this structure.
};


//
// Global variables.
//
extern BYTE DebugLevel;
extern BOOL StartGPSSyncThread;
extern DWORD SysClkSyncThreadID;


//
// Function prototyping.
//
DWORD WINAPI ReaderThreadSeisAD18(PVOID);
BOOL CALLBACK ConfigDlgProcSeisAD18(HWND, UINT, WPARAM, LPARAM);
BOOL ReadStartOfSecond(HANDLE, SYSTEMTIME*, PBYTE, PDWORD);
VOID Adjust2NearestSec(SYSTEMTIME*);
VOID ConfigureSeisAD18(HANDLE, BYTE);
BOOL SeisAD18ConfigOK(BYTE, BYTE);



/*************************************************************************/
/*  Manager thread for device Geosig SeisAD18. DeviceID=210.             */
/*************************************************************************/
DWORD WINAPI SeisAD18MngrThrd(PVOID StartupInfo) 
{
   MSG msg;
   int Status;
   BYTE Index, Gain;
   TCHAR ConfigFile[MAX_PATH+1], ChNr, UniqueDeviceName[65], Message[100];
   TCHAR DeviceResources[256], str10[11], BaudRate[7];
   WORD BlockSize, NumBlocksPrChannel, SampleRate;
   BOOL SerialRead, MemAlloc, ConfigRead;
   HANDLE hReaderThread, hComm;
   STARTINFO_DEVMANAGER SI;
   STARTINFO_DEVREADER StartInfo;
   DEVICE_CONFIG DeviceConf, DevConfDlg;
   DEVMNGR_INTERF_STRUCT DS;
   DWORD ParentThreadID, CurrentThreadID, ReaderThreadID;
   DWORD ChannelSize, status, BlockNr;
   MSG_PARAM ID, OD;
   MsgList NotifyList;

   
   //
   // Perform thread initialization.
   //
   hReaderThread=NULL;                               // The serial-reader thread is not running.
   CopyMemory((PVOID)&SI, StartupInfo, sizeof(SI));  // Copy startup information to local buffer.
   ParentThreadID=SI.ThreadID;                       // ThreadID for parent thread.
   CurrentThreadID=GetCurrentThreadId();             // ThreadID for this thread.
   

   //
   // Handle device configuration.
   //
   ConfigRead=FALSE;
   wcscpy(ConfigFile, SI.DeviceDir); 
   wcscat(ConfigFile, TEXT("\\device.cnf"));
   if (ExistFile(SI.DeviceDir, TEXT("device.cnf"))) {
      // Read device config from config file.
      ConfigRead=ReadFileToMemory(ConfigFile, (PVOID)&DeviceConf, (DWORD)sizeof(DeviceConf));
   }
   if (!ConfigRead) {
      // A valid config file could not be read.
      // Set up a default digitizer configuration.
      DeviceConf.DeviceID=210;               // SeisAD18 uses device-id 210.
      wcscpy(DeviceConf.PortName, TEXT("COM1"));   // Assume digitizer on COM1.
      DeviceConf.NumChannels=3;              // Assume digitizer has been configured for 3 channels.
      DeviceConf.ConfigNr=17;                // Assume 19200 baud, 100hz, gain=4.
      DeviceConf.TimeCorrection=0.0;         // Default time correction (seconds).

      // Write settings to config file.
      WriteFileFromMemory(ConfigFile, (PVOID)&DeviceConf, (DWORD)sizeof(DeviceConf));
   }


   // 
   // Create a unique device instance name.
   //
   wcscpy(UniqueDeviceName, TEXT("SeisAD18 at "));
   wcscat(UniqueDeviceName, DeviceConf.PortName);

   //
   // Register this process in the process database.
   //
   if (!SI.ConfigOnly) RegisterProcess(UniqueDeviceName, 1, CurrentThreadID);

   //
   // Create device resource string.
   // 
   wcscpy(DeviceResources, TEXT("Serial port: "));
   wcscat(DeviceResources, DeviceConf.PortName);

   //
   // Find sample-rate, communication speed and Gain.
   //
   if (DeviceConf.ConfigNr== 1) { SampleRate=25;  wcscpy(BaudRate,TEXT("9600"));   Gain=1;  } 
   if (DeviceConf.ConfigNr== 2) { SampleRate=25;  wcscpy(BaudRate,TEXT("9600"));   Gain=4;  }
   if (DeviceConf.ConfigNr== 3) { SampleRate=25;  wcscpy(BaudRate,TEXT("9600"));   Gain=16; }
   if (DeviceConf.ConfigNr== 4) { SampleRate=25;  wcscpy(BaudRate,TEXT("19200"));  Gain=1;  }
   if (DeviceConf.ConfigNr== 5) { SampleRate=25;  wcscpy(BaudRate,TEXT("19200"));  Gain=4;  }
   if (DeviceConf.ConfigNr== 6) { SampleRate=25;  wcscpy(BaudRate,TEXT("19200"));  Gain=16; }
   if (DeviceConf.ConfigNr== 7) { SampleRate=50;  wcscpy(BaudRate,TEXT("9600"));   Gain=1;  }
   if (DeviceConf.ConfigNr== 8) { SampleRate=50;  wcscpy(BaudRate,TEXT("9600"));   Gain=4;  }
   if (DeviceConf.ConfigNr== 9) { SampleRate=50;  wcscpy(BaudRate,TEXT("9600"));   Gain=16; }
   if (DeviceConf.ConfigNr==10) { SampleRate=50;  wcscpy(BaudRate,TEXT("19200"));  Gain=1;  }
   if (DeviceConf.ConfigNr==11) { SampleRate=50;  wcscpy(BaudRate,TEXT("19200"));  Gain=4;  }
   if (DeviceConf.ConfigNr==12) { SampleRate=50;  wcscpy(BaudRate,TEXT("19200"));  Gain=16; }
   if (DeviceConf.ConfigNr==13) { SampleRate=50;  wcscpy(BaudRate,TEXT("115200")); Gain=1;  }
   if (DeviceConf.ConfigNr==14) { SampleRate=50;  wcscpy(BaudRate,TEXT("115200")); Gain=4;  }
   if (DeviceConf.ConfigNr==15) { SampleRate=50;  wcscpy(BaudRate,TEXT("115200")); Gain=16; }
   if (DeviceConf.ConfigNr==16) { SampleRate=100; wcscpy(BaudRate,TEXT("19200"));  Gain=1;  }
   if (DeviceConf.ConfigNr==17) { SampleRate=100; wcscpy(BaudRate,TEXT("19200"));  Gain=4;  }
   if (DeviceConf.ConfigNr==18) { SampleRate=100; wcscpy(BaudRate,TEXT("19200"));  Gain=16; }
   if (DeviceConf.ConfigNr==19) { SampleRate=100; wcscpy(BaudRate,TEXT("115200")); Gain=1;  }
   if (DeviceConf.ConfigNr==20) { SampleRate=100; wcscpy(BaudRate,TEXT("115200")); Gain=4;  }
   if (DeviceConf.ConfigNr==21) { SampleRate=100; wcscpy(BaudRate,TEXT("115200")); Gain=16; }
   if (DeviceConf.ConfigNr==22) { SampleRate=200; wcscpy(BaudRate,TEXT("115200")); Gain=1;  }
   if (DeviceConf.ConfigNr==23) { SampleRate=200; wcscpy(BaudRate,TEXT("115200")); Gain=4;  }
   if (DeviceConf.ConfigNr==24) { SampleRate=200; wcscpy(BaudRate,TEXT("115200")); Gain=16; }


   //
   // Calculate blocksize, channelsize and number of blocks pr. channel.
   //
   NumBlocksPrChannel=60; 
   BlockSize=(WORD)(sizeof(BLOCK_HEADER)+(SampleRate*4));
   ChannelSize=NumBlocksPrChannel*BlockSize;


   // Open the serial port.
   if (!SI.ConfigOnly) {
#ifdef WINCE
	  // All ports in Windows CE must end with ':'
	  TCHAR CEPortName[7];
	  wcscpy(CEPortName, DeviceConf.PortName);
	  wcscat(CEPortName, TEXT(":"));
	  hComm=CreateFile(CEPortName, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
#else
	  TCHAR SerPort[10]
      wcscpy(SerPort, TEXT("\\\\.\\")); wcscat(SerPort, DeviceConf.PortName);
      hComm=CreateFile(SerPort, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
#endif
	  if (hComm==INVALID_HANDLE_VALUE) {
         // Update program log.
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(": Unable to open communication port "));
         wcscat(Message, DeviceConf.PortName);
         wcscat(Message, TEXT("."));
         WriteSystemLog(Message, 3);
      }
   }

   
   //
   // Initialize the interface structure.
   //
   // Insert manufacturer name and model.
   wcscpy(DS.Manufacturer, TEXT("SeisProd"));
   wcscpy(DS.Model, TEXT("SEISAD18"));

   // Insert samplerate.
   DS.SampleRate=SampleRate;

   // Insert number of samples in each block.
   DS.SamplesPrBlock=SampleRate;

   // Insert number of channels
   DS.NumChannels=DeviceConf.NumChannels;

   // Insert block size, block header included.
   DS.BlockSize=BlockSize;

   // Insert block size measured in milliseconds.
   DS.BlockSizeInMS=1000;

   // Insert number of blocks pr channel.
   DS.NumBlocksPrChannel=NumBlocksPrChannel;

   // Make sure all pointers to channel data is NULL.
   for (ChNr=0;ChNr<MAX_CHANNELS_PR_DEVICE;ChNr++) DS.ChannelData[ChNr]=(PVOID)NULL;
   
   
   //
   // Allocate memory for the channel buffers, and set up the pointers to the buffers.
   //
   MemAlloc=TRUE;
   for (Index=0;Index<DeviceConf.NumChannels;Index++) {
      DS.ChannelData[Index]=(PVOID)LocalAlloc(LPTR, (UINT)ChannelSize);
      if (DS.ChannelData[Index]==NULL) MemAlloc=FALSE;
   }
   // Check that memory allocation succeded.
   if (!MemAlloc) {
         // Update program log.
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(":"));
         wcscat(Message, TEXT(" Could not allocate memory."));
         WriteSystemLog(Message, 3);
   }
   

   //
   // If running in debug mode, write a message to the log.
   //
   if (DebugLevel) {
      wcscpy(Message, UniqueDeviceName);
      wcscat(Message, TEXT(": Thread started."));
      WriteSystemLog(Message, 1);
   }
   

   //
   // Signal that the thread is started.
   //
   SetEvent(SI.StartEvent);


   //
   // Enter the thread message loop. The thread will now start to process messages.
   //
   while (GetMessage(&msg, (HWND)NULL, 0, 0)) 
   { 
   
      // If debuglevel is 2 then log all messages recieved.
      if (DebugLevel>1) {
         TCHAR str10[11];
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(": Recieved message "));
         _itow(msg.message, str10, 10);
         wcscat(Message, str10);
         WriteSystemLog(Message, 1);
      }

      // Save message parameters.
      ZeroMemory((PVOID)&ID, (DWORD)sizeof(ID));
      if (msg.lParam) ID=*((MSG_PARAM*)msg.lParam);
      
      // Initialize return-message parameter to default values.
      ZeroMemory((PVOID)&OD, (DWORD)sizeof(OD));
      OD.bStatus=TRUE;
      OD.idThread=CurrentThreadID;


      // Process message.
      switch (msg.message) {
         case TM_START:                            
         //
         // This message is sendt from a higher level 
         // thread to start reading the serial port.
         //
         if (hReaderThread==NULL && hComm!=INVALID_HANDLE_VALUE && MemAlloc && !SI.ConfigOnly) {
            if (SysClkSyncThreadID) {
               // The system clock is synchronized by some thread.
               // Subscribe to the TM_TIMESYNC_STATUS message from thread.
               Subscribe(TM_TIMESYNC_STATUS, SysClkSyncThreadID, CurrentThreadID, (HWND)NULL);

               // Get the current synchronization status.
               StartInfo.SyncStatus=GetCurrentTimeSyncStatus(SysClkSyncThreadID);
            } else {
               // The system clock is not synchronized and
               // there is no other synchronization possibilities.
               StartInfo.SyncStatus=0; 
            }

            // Set up startup information for the serial reader thread routine.
            SerialRead=TRUE;
            StartInfo.hComm=hComm;
            wcscpy(StartInfo.CommSpeed, BaudRate);
            StartInfo.ThreadID=CurrentThreadID;
            StartInfo.SerialRead=&SerialRead;
            StartInfo.DS=&DS;
            StartInfo.ConfigNr=DeviceConf.ConfigNr;
            StartInfo.BlockNr=&BlockNr;
            StartInfo.TimeCorrection=DeviceConf.TimeCorrection;
            StartInfo.hSyncStatusMutex=CreateMutex(NULL, FALSE, NULL); 

            // Start the thread.
            hReaderThread=CreateThread(NULL, 0, ReaderThreadSeisAD18, (VOID*)&StartInfo, 0, &ReaderThreadID);
         }

         // Set return values. 
         if (!hReaderThread) {
            OD.bStatus=FALSE;
            OD.wStatus=1;
         }
         break;


         
         case TM_STOP:
         //
         // This message is sendt from a higher level thread to stop
         // reading the serial port. Returns the default return values.
         //
         if (hReaderThread) {
            // Make the serial thread loop exit, thereby stopping the thread.
            SerialRead=FALSE;

            // Wait up to 2 seconds for the thread to stop.
            status=WaitForSingleObject(hReaderThread, 2000);
            if (status==WAIT_TIMEOUT) TerminateThread(hReaderThread, 0);

            // Unsubscribe to the TM_TIMESYNC_STATUS message as needed.
            if (SysClkSyncThreadID) UnSubscribe(TM_TIMESYNC_STATUS, SysClkSyncThreadID, CurrentThreadID, NULL);

            // Close the mutex object.
            CloseHandle(StartInfo.hSyncStatusMutex);
         
            // The thread is now stopped, invalidate the thread handle.
				CloseHandle(hReaderThread);
            hReaderThread=NULL;
         }
         break;


         case TM_CONFIGURE:            
         //
         // This message is sendt from a higher level thread
         // when it wants to configure the digitizer settings.
         //
         if (hReaderThread==NULL) {
            // Let user work on a copy of the current device settings.
            DevConfDlg=DeviceConf;

            // Display the dialog box.
            Status=DialogBoxParam(SI.hInstance, MAKEINTRESOURCE(IDD_CNF_SEISAD18), NULL, 
                                  (DLGPROC)ConfigDlgProcSeisAD18, (LPARAM)&DevConfDlg);  

            // Save settings if requested.
            if (Status==1) {
               // Save the new configuration.
               DeviceConf=DevConfDlg;

               // Write configuration to disk.
               WriteFileFromMemory(ConfigFile, (PVOID)&DeviceConf, (DWORD)sizeof(DEVICE_CONFIG));

               // Create a unique device instance name.
               wcscpy(UniqueDeviceName, TEXT("SeisAD18 at "));
               wcscat(UniqueDeviceName, DeviceConf.PortName);
            }

            // Set return values. 
            if (Status==0) {
               OD.bStatus=FALSE;
               OD.wStatus=100;
            }
         }
         break;

         case TM_SHOW_CONFIG:
            //
            // When this message is received, dump all configuration
            // parameters not given in the device resource string to 
            // the system log.
            //
            wcscpy(Message, UniqueDeviceName);
            wcscat(Message, TEXT(": Time Correction="));
            swprintf(str10,TEXT("%.3f"), DeviceConf.TimeCorrection);
            wcscat(Message, str10);
            WriteSystemLog(Message, 1);
            wcscpy(Message, UniqueDeviceName);
            wcscat(Message, TEXT(": Communication speed="));
            wsprintf(str10,TEXT("%s"), BaudRate);
            wcscat(Message, str10);
            WriteSystemLog(Message, 1);
            wcscpy(Message, UniqueDeviceName);
            wcscat(Message, TEXT(": Gain="));
            wsprintf(str10,TEXT("%d"), Gain);
            wcscat(Message, str10);
            WriteSystemLog(Message, 1);
         break;


         case TM_GET_INTERFACE:
         //
         // This message is used to get a pointer to the interface structure.
         //
         OD.vPointer=(PVOID)&DS;
         break;   


         case TM_GET_UNIQUE_DEVICE_NAME:
         //
         // This message is used to get a string, 
         // identifying this unique device instance.
         //
         wcscpy(OD.cStatus, UniqueDeviceName);
         break;   


         case TM_GET_RESOURCE_STRING:
         //
         // This message is used to get a string, 
         // identifying the system resources used by this device.
         //
         wcscpy(OD.cStatus, DeviceResources);
         break;   


         case TM_SUBSCRIBE:
         //
         // This message is sendt from a window or thread
         // Who wants to be informed about an event.
         // This message must be accompanied by a MSG_PARAM structure.
         // The MSG_PARAM structure must have a pointer to a valid
         // SUBSCRIBE structure.
         //
         //
         // Add window/thread to notify list.
         if (ID.vPointer) if (!NotifyList.Subscribe((SUBSCRIBE*)ID.vPointer)) {
            OD.bStatus=FALSE;
            OD.wStatus=3;
         }
         break;   


         case TM_UNSUBSCRIBE:
         //
         // This message is sendt from a window or thread who
         // wants to remove a notification from the notify list.
         // This message must be accompanied by a MSG_PARAM structure.
         // The MSG_PARAM structure must have a pointer to a valid
         // SUBSCRIBE structure.
         //
         //
         // Remove notification from list.
         if (ID.vPointer) NotifyList.UnSubscribe((SUBSCRIBE*)ID.vPointer);
         break;   


         case TM_TIMESYNC_STATUS:
         //
         // This message is sendt from the time synchronization
         // thread to inform about a change in the synchronization
         // status.
         WaitForSingleObject(StartInfo.hSyncStatusMutex, INFINITE); 
         StartInfo.SyncStatus=(BYTE)msg.wParam;
         ReleaseMutex(StartInfo.hSyncStatusMutex);
         break;

         
         case TM_EXIT:                 
         //
         // This is a request for the thread to end.
         // Close down all processes that are depandent on this thread.
         // The 'StopDependantThreads' function will return immediatly,
         // and will send a 'TM_DEPTHREADS_STOPPED' message back when
         // all dependant threads have been stopped.
         /*if (!hReaderThread)*/ StopDependantThreads();
         break;


         case TM_DEPTHREADS_STOPPED:
         //
         // This message indicates that all dependent threads have been stopped.
         // The thread will cleanup and exit when this message is recieved.         
         //


		 // Next if statement Added 04.11.2003. Differs from PC version.

         // First stop the device-reader thread.
         if (hReaderThread) {
            SerialRead=FALSE;                                            // Signal the reader thread.
            status=WaitForSingleObject(hReaderThread, 2000);             // Wait for the thread to stop.
            if (status==WAIT_TIMEOUT) TerminateThread(hReaderThread, 0); // If thread wont stop, kill it.
            hReaderThread=NULL;
         }

		 //


         if (!SI.ConfigOnly) {
            // Unregister this process from the process database.
            UnregisterProcess(CurrentThreadID);

            // Close the communication port.
            CloseHandle(hComm);
         }

         // Free memory buffers.
         for (Index=0;Index<DeviceConf.NumChannels;Index++) {
            if (DS.ChannelData[Index]) LocalFree((HLOCAL)DS.ChannelData[Index]); 
         }

         // End the message loop.
         ClearMessageQueue(); // Delete all messages in message queue.
         PostQuitMessage(0);  // Post WM_QUIT message.   
         break;
   
         
         //
         // Handle messages from device reader thread.
         //
         case TM_BUFFERUPDATED:  // A new block of data has been inserted into buffer.
         // Send Notification to all subscribed threads/windows.
         // The notofication is accompanied with this thread's thread-ID, and the blocknr of the updated block.
         NotifyList.Notify((UINT)TM_BUFFERUPDATED, (WPARAM)CurrentThreadID, (LPARAM)BlockNr);
         break;


         case TM_DEVICEREADER_NODATAINPUT:   
         //
         // The device reader has lost data input.
         //
         // Send Notification to all subscribed threads/windows.
         NotifyList.Notify((UINT)TM_DEVICEREADER_NODATAINPUT, WPARAM(CurrentThreadID), 0);

         // Write a message to the log.
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(": Not receiving data."));
         WriteSystemLog(Message, 2);
         break;


         case TM_DEVICEREADER_DATAINPUTOK:   
         //
         // The device reader has regained data input.
         //
         // Send Notification to all subscribed threads/windows.
         NotifyList.Notify((UINT)TM_DEVICEREADER_DATAINPUTOK, WPARAM(CurrentThreadID), 0);

         // Write a message to the log.
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(": Receiving data."));
         WriteSystemLog(Message, 1);
         break;


         case TM_DEVICEREADER_BADBLOCK:   // A bad block of data was received from the device.
         // Send Notification to all subscribed threads/windows.
         // The notofication is accompanied with this thread's thread-ID, and the blocknr of the bad block.
         NotifyList.Notify((UINT)TM_DEVICEREADER_BADBLOCK, (WPARAM)CurrentThreadID, (LPARAM)BlockNr);

         // Write a message to the log.
         wcscpy(Message, UniqueDeviceName);
         wcscat(Message, TEXT(": Received a bad block from device."));
         WriteSystemLog(Message, 2);
         break;
      }
      
      // If this is requested by the thread who sent the message
      // then return status values, and release the caller thread.
      if (((MSG_PARAM*)msg.lParam)!=NULL && msg.wParam!=NULL) {
         *(MSG_PARAM*)msg.lParam=OD;
         SetEvent((HANDLE)msg.wParam);
      }
   } 


   // Free up memeory allocated in the MsgList class object.
   NotifyList.~MsgList();

   // If running in debug mode, write a message to the log.
   if (DebugLevel) {
      wcscpy(Message, UniqueDeviceName);
      wcscat(Message, TEXT(": Thread stopped."));
      WriteSystemLog(Message, 1);
   }

   // Exit this thread
   ExitThread(0);
   return 0L;
   
}




/*************************************************************************/
/* Serial port reader procedure for the Geosig SeisAD18 device.          */
/*************************************************************************/
DWORD WINAPI ReaderThreadSeisAD18(PVOID StartupStruct)
{
   DCB dcb;
   BYTE DigNr, ChnlNr, *pDataBlocks, GPSStatus;
   BYTE Buffer[13], *bPtr, BuffNr, NumDig;
   BYTE ChkByte1, ChkByte2, ChkByte3;
   BOOL DataLost, ReadStatus;
   WORD BlockSize, NumChannels, NumBlocksPrChannel;
   WORD SampleRate, BlocksPrSec, BlocksRead;
   DWORD BytesRead, BlockNr, OffsetToBlockData;
   SYSTEMTIME ReadTime, ReadTime1, ReadTime2;
   COMMTIMEOUTS ct={0, 0, 800, 0, 0};
   SeisAD18Buffer InBuff1, InBuff2;
   STARTINFO_DEVREADER *SI;
   BLOCK_HEADER DBH;
   BOOL ReadSOS, BufferFull, CheckSumOK;


   // Copy values from startup structure.
   SI=(STARTINFO_DEVREADER*)StartupStruct;         // Use pointer of correct type.
   NumChannels=SI->DS->NumChannels;                // Number of channels digitizer is using.
   SampleRate=SI->DS->SampleRate;
   NumBlocksPrChannel=SI->DS->NumBlocksPrChannel;  // Number of 1 sec. blocks in buffer.
   BlockSize=SI->DS->BlockSize;                    // Size of 1 sec. block header included
   NumDig=NumChannels/3;

   //Initialize the communications parameters. 
   SetupComm(SI->hComm, 1024, 0); 
   GetCommState(SI->hComm, &dcb);
#ifdef WINCE
   dcb.BaudRate     = _wtoi(SI->CommSpeed);
   dcb.Parity       = NOPARITY;
   dcb.StopBits     = ONESTOPBIT;
   dcb.ByteSize     = 8;
#else
   TCHAR DevCntrlStr[64];
   wcscpy(DevCntrlStr, TEXT("baud="));
   wcscat(DevCntrlStr, SI->CommSpeed);
   wcscat(DevCntrlStr, TEXT(" parity=n data=8 stop=1"));
   BuildCommDCB(DevCntrlStr, &dcb);
#endif
   dcb.fAbortOnError=FALSE; 
   SetCommState(SI->hComm, &dcb);
   SetCommTimeouts(SI->hComm, &ct);

   // Configure device. This is done by sending
   // a series of NULL characters to the device.
   ConfigureSeisAD18(SI->hComm, SI->ConfigNr);

   // Clear the input buffer.
   PurgeComm(SI->hComm, PURGE_RXCLEAR);

   // Initiate the instances of the buffer classes.
   InBuff1.InitBuffer((BYTE)NumChannels, SampleRate);
   InBuff2.InitBuffer((BYTE)NumChannels, SampleRate);

   // Calculate number of 12 byte blocks pr second.
   BlocksPrSec=(NumChannels/3)*SampleRate;

   // Start reading the serial port.
   BuffNr=1;
   BlockNr=0;
   ReadSOS=DataLost=TRUE;
   while(*SI->SerialRead) {

      //
      // Read 12 bytes into buffer.
      //
      if (ReadSOS) {
         // Read the 12 first bytes of the next second into buffer.
         ReadStatus=ReadStartOfSecond(SI->hComm, &ReadTime, Buffer, &BytesRead);
         if (BytesRead==12) {
            // Use readfile function for the rest of this second.
            ReadSOS=FALSE; BlocksRead=0;
            // Save the time when the first bytes was read.
            if (BuffNr==1) ReadTime1=ReadTime;
            if (BuffNr==2) ReadTime2=ReadTime;
         }
      } else {
         // Read the next 12 bytes into buffer.
         ReadStatus=ReadFile(SI->hComm, (LPVOID)Buffer, 12, &BytesRead, NULL);
      }

      //
      // Handle the 12 byte block.
      //
      if (BytesRead==12 && !ReadSOS) {
         // Update 12 byte-block counter.
         BlocksRead++;

         // Insert data into current buffer.
         if (BuffNr==1) InBuff1.InsertBuffer(Buffer);
         if (BuffNr==2) InBuff2.InsertBuffer(Buffer);

         // Check and copy data for previous second 
         // if we have received the checksum bytes.
         if (BlocksRead==20) {
            // Check that the other buffer has data.
            if (BuffNr==1) BufferFull=InBuff2.IsBufferFull();
            if (BuffNr==2) BufferFull=InBuff1.IsBufferFull();
            if (BufferFull) {
               // Create block header.
               DBH.BlockID=BlockNr;
               DBH.NullData=FALSE;
               DBH.DataLost=DataLost;
               DBH.EventDetected=FALSE;
               DBH.EventTime=0;
               for (ChnlNr=0;ChnlNr<NumChannels;ChnlNr++) {
                  bPtr=((BYTE*)SI->DS->ChannelData[ChnlNr])+(BlockSize*BlockNr);
                  CopyMemory((VOID*)bPtr, (VOID*)&DBH, (DWORD)sizeof(BLOCK_HEADER));
               }
            
               // For all digitizers, copy data for all channels
               // into the device manager buffers. If data is not
               // OK, then zero's are copied to the buffers instead.
               for (DigNr=0;DigNr<NumDig;DigNr++) {
                  // Use correct block time. If GPS status is 0
                  // then the samples are synchronized to start of
                  // second. If GPS status is 1, then we must use
                  // system clock for timestamping.
                  if (BuffNr==1) {
                     DBH.TimeTag=ReadTime2;
                     GPSStatus=InBuff2.GetGPSStatus(DigNr);
                  } else {
                     DBH.TimeTag=ReadTime1;
                     GPSStatus=InBuff1.GetGPSStatus(DigNr);
                  }
                  if (!GPSStatus) Adjust2NearestSec(&DBH.TimeTag);
                  // Adjust time as requested in setup.
                  if (SI->TimeCorrection>0.0) AddToTime(&DBH.TimeTag, SI->TimeCorrection);
                  if (SI->TimeCorrection<0.0) SubstractFromTime(&DBH.TimeTag, -SI->TimeCorrection);
                  
                  // Find the timequality.
                  if (GPSStatus) {
                     // Device has no PPS signal.
                     if (SysClkSyncThreadID) {
                        // The system clock is synchronized. 
                        WaitForSingleObject(SI->hSyncStatusMutex, INFINITE); 
                        if (SI->SyncStatus==0) DBH.TimeQuality=2;
                        if (SI->SyncStatus==1) DBH.TimeQuality=1;
                        if (SI->SyncStatus==2) DBH.TimeQuality=3;
                        ReleaseMutex(SI->hSyncStatusMutex);
                     } else {
                        // The system clock is NOT synchronized. 
                        DBH.TimeQuality=0; 
                     }
                  } else {
                     // Device has PPS signal.
                     DBH.TimeQuality=2;
                  }

                  // Verify checksum for this digitizer.
                  if (BuffNr==1) {
                     InBuff1.GetChkSum(DigNr, &ChkByte1, &ChkByte2, &ChkByte3);
                     CheckSumOK=InBuff2.VerifyChkSum(DigNr, ChkByte1, ChkByte2, ChkByte3);
                  } else {
                     InBuff2.GetChkSum(DigNr, &ChkByte1, &ChkByte2, &ChkByte3);
                     CheckSumOK=InBuff1.VerifyChkSum(DigNr, ChkByte1, ChkByte2, ChkByte3);
                  }
                  
                  // If checksum for this digitizer is OK, then copy samples
                  // from this digitizer's channels into device manager buffers. 
                  if (CheckSumOK) {
                     // The samples from this digitizer seems to be OK.
                     // For all 3 channels, copy samples to DM buffer.
                     OffsetToBlockData=(BlockNr*BlockSize)+sizeof(BLOCK_HEADER);
                     for (ChnlNr=DigNr*3;ChnlNr<=(DigNr*3+2);ChnlNr++) {
                        // Copy samples for this channel.
                        pDataBlocks=(PBYTE)SI->DS->ChannelData[ChnlNr];
                        if (BuffNr==1) InBuff2.CopyChnlData(ChnlNr, pDataBlocks+OffsetToBlockData);
                        if (BuffNr==2) InBuff1.CopyChnlData(ChnlNr, pDataBlocks+OffsetToBlockData);
                        // Copy the block header for this channel.
                        bPtr=((BYTE*)SI->DS->ChannelData[ChnlNr])+(BlockSize*BlockNr);
                        CopyMemory((VOID*)bPtr, (VOID*)&DBH, (DWORD)sizeof(BLOCK_HEADER));
                     }
                  } else {
                     // The samples from this digitizer are bad.
                     // Fill DM buffer for theese 3 channels with zeors.
                     OffsetToBlockData=(BlockNr*BlockSize)+sizeof(BLOCK_HEADER);
                     for (ChnlNr=DigNr*3;ChnlNr<=DigNr*3+2;ChnlNr++) {
                        // Fill channel with nulldata.
                        pDataBlocks=(PBYTE)SI->DS->ChannelData[ChnlNr];
                        ZeroMemory((PVOID)(pDataBlocks+OffsetToBlockData), SampleRate*4);
                        // Copy the block header for this channel.
                        DBH.NullData=TRUE;
                        bPtr=((BYTE*)SI->DS->ChannelData[ChnlNr])+(BlockSize*BlockNr);
                        CopyMemory((VOID*)bPtr, (VOID*)&DBH, (DWORD)sizeof(BLOCK_HEADER));
                     }
                     // Inform parent thread that checksum failed.
                     PostThreadMessage(SI->ThreadID, TM_DEVICEREADER_BADBLOCK, 0, 0); 
                  }
               }

               // Clear the buffer that heldt the previous second.
               if (BuffNr==1) InBuff2.ResetBuffer();
               if (BuffNr==2) InBuff1.ResetBuffer();

               // If data stream has just been restored,then inform parent thread.
               if (DataLost) {
                  DataLost=FALSE;
                  PostThreadMessage(SI->ThreadID, TM_DEVICEREADER_DATAINPUTOK, 0, 0);
               }

               // Return block nr for updated block.
               *SI->BlockNr=BlockNr;

               // Signal parent process that buffer is updated.
               PostThreadMessage(SI->ThreadID, TM_BUFFERUPDATED, 0, 0);

               // Advance to the next block.
               BlockNr++;
               if (BlockNr==NumBlocksPrChannel) BlockNr=0;
            }
         }

         // This buffer is full now. 
         // Next second goes into the other buffer.
         if (BlocksRead==BlocksPrSec) {
            // Read start of second next.
            // This will verify that we are in sync with the device.
            ReadSOS=TRUE;     
            // Switch to the other buffer.
            if (BuffNr==1) { BuffNr=2; } else { BuffNr=1; }
         }

      } else {
         // Failed to read 12 bytes.The digitizer has been
         // disconnected, or another comm error has occured.
         if (DataLost==FALSE) {
            ReadSOS=TRUE; // We must resync with start of second block.
            DataLost=TRUE;
            BuffNr=1; InBuff1.ResetBuffer(); InBuff2.ResetBuffer();
            PostThreadMessage(SI->ThreadID, TM_DEVICEREADER_NODATAINPUT, 0, 0); 
         }

         // Try to configure the device.
         ConfigureSeisAD18(SI->hComm, SI->ConfigNr);

         // Wait before trying a new read.
         if (!ReadStatus) Sleep(500);
      }
   }
   
   ExitThread(0);
   return 0L;
   
}



/***********************************************************************/
/* Searches for and reads start of second (12 bytes) from digitizer 0. *
/***********************************************************************/
BOOL ReadStartOfSecond(HANDLE hComm, SYSTEMTIME* ReadTime,
                       PBYTE Buffer, PDWORD BytesRead)
{
   BYTE Char;
   DWORD NumBytes;
   BOOL ReadStatus, Found=FALSE;
   
   
   // Find the 3 leading 0x55 bytes.
   *BytesRead=0;
   while (!Found) {
      ReadStatus=ReadFile(hComm, (LPVOID)&Char, 1, &NumBytes, NULL);
      if (!NumBytes) return ReadStatus;
      if (Char!=0x55) *BytesRead=0;
      if (Char==0x55) (*BytesRead)++;
      if (*BytesRead==3) { 
         Found=TRUE;
         GetSystemTime(ReadTime);
      }
   }

   // Read the rest of the string.
   Buffer[0]=Buffer[1]=Buffer[2]=0x55;
   ReadStatus=ReadFile(hComm, (LPVOID)(Buffer+3), 9, &NumBytes, NULL);
   *BytesRead=*BytesRead+NumBytes;

   return ReadStatus;
}






/*******************************************************************/
/*    Configuration dialog box procedure for SeisAD18.             */
/*******************************************************************/
BOOL CALLBACK ConfigDlgProcSeisAD18(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
   SHINITDLGINFO shidi; // New in PocketPC version
   HWND hwndDlgItem;
   DWORD dwIndex;
   INT Result, Index;
   static DEVICE_CONFIG *DC;
   TCHAR PortNr, Str20[21], Message[100];
	TCHAR ConfTab[24][64]={TEXT("Baud=9600, Fs=25, Gain=1"),
				   		    TEXT("Baud=9600, Fs=25, Gain=4"),
							    TEXT("Baud=9600, Fs=25, Gain=16"),
								 TEXT("Baud=19200, Fs=25, Gain=1"),
								 TEXT("Baud=19200, Fs=25, Gain=4"),
								 TEXT("Baud=19200, Fs=25, Gain=16"),
								 TEXT("Baud=9600, Fs=50, Gain=1"),
								 TEXT("Baud=9600, Fs=50, Gain=4"),
								 TEXT("Baud=9600, Fs=50, Gain=16"),
								 TEXT("Baud=19200, Fs=50, Gain=1"),
								 TEXT("Baud=19200, Fs=50, Gain=4"),
								 TEXT("Baud=19200, Fs=50, Gain=16"),
								 TEXT("Baud=115200, Fs=50, Gain=1"),
								 TEXT("Baud=115200, Fs=50, Gain=4"),
								 TEXT("Baud=115200, Fs=50, Gain=16"),
								 TEXT("Baud=19200, Fs=100, Gain=1"),
								 TEXT("Baud=19200, Fs=100, Gain=4"),
								 TEXT("Baud=19200, Fs=100, Gain=16"),
								 TEXT("Baud=115200, Fs=100, Gain=1"),
								 TEXT("Baud=115200, Fs=100, Gain=4"),
								 TEXT("Baud=115200, Fs=100, Gain=16"),
								 TEXT("Baud=115200, Fs=200, Gain=1"),
								 TEXT("Baud=115200, Fs=200, Gain=4"),
								 TEXT("Baud=115200, Fs=200, Gain=16")};
   TCHAR Ports[16][6]={TEXT("COM1"), TEXT("COM2"), TEXT("COM3"), TEXT("COM4"), TEXT("COM5"), TEXT("COM6"),
		                TEXT("COM7"), TEXT("COM8"), TEXT("COM9"), TEXT("COM10"),TEXT("COM11"),TEXT("COM12"),
							 TEXT("COM13"),TEXT("COM14"),TEXT("COM15"),TEXT("COM16")};

   switch(message)  {
      case WM_INITDIALOG:

 	  // Create a Done button and size it. New in PocketPC version.
 	  shidi.dwMask = SHIDIM_FLAGS;
	  shidi.dwFlags = SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
	  shidi.hDlg = hWndDlg;
	  SHInitDialog(&shidi);
      
	  // Create pointer to the device config structure.
      DC=(DEVICE_CONFIG*)lParam;
   
      // Initialize the 'Device is connected to' combo box.
      hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO1);
      for (PortNr=0;PortNr<16;PortNr++) {
         dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)Ports[PortNr]);
         SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, PortNr);
         if (!wcscmp(DC->PortName, Ports[PortNr])) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);
      }

      // Initialize the 'number of channels' combo box.
      hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO2);
      dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR) TEXT("3"));
      SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, 3);
      if (DC->NumChannels==3) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);
      dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR) TEXT("6"));
      SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, 6);
      if (DC->NumChannels==6) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);
      dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR) TEXT("9"));
      SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, 9);
      if (DC->NumChannels==9) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);
      dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR) TEXT("12"));
      SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, 12);
      if (DC->NumChannels==12) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);

      // Initialize the operating configuration combo box.
      hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO3);
      for (Index=0;Index<24;Index++) {
         dwIndex=SendMessage(hwndDlgItem, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)ConfTab[Index]);
         SendMessage(hwndDlgItem, CB_SETITEMDATA, dwIndex, Index+1);
         if (DC->ConfigNr==Index+1) SendMessage(hwndDlgItem, CB_SETCURSEL, dwIndex, 0);
      }
      
      // Initialize the 'time correction (seconds)' edit box.
      swprintf(Str20, TEXT("%.3f"), DC->TimeCorrection);
      SetDlgItemText(hWndDlg, IDC_TIME_CORR, Str20);

      // Bring window to front.
      // BringWindowToTop(hWndDlg);
	  SetForegroundWindow(hWndDlg);

      return TRUE;


      case WM_COMMAND:
      switch(LOWORD(wParam)) {
         case IDOK:  // User has pressed the OK button.
         // Read current choice from the 'Device is connected to' combo box.
         hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO1);
         dwIndex=SendMessage(hwndDlgItem, CB_GETCURSEL, 0, 0);
         SendMessage(hwndDlgItem, CB_GETLBTEXT, dwIndex, (LPARAM)&DC->PortName);

         // Read current choice from the 'Number of channels' combo box.
         hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO2);
         dwIndex=SendMessage(hwndDlgItem, CB_GETCURSEL, 0, 0);
         DC->NumChannels=(BYTE)SendMessage(hwndDlgItem, CB_GETITEMDATA, dwIndex, 0);

         // Read current choice from the operating configuration combo box.
         hwndDlgItem=GetDlgItem(hWndDlg, IDC_COMBO3);
         dwIndex=SendMessage(hwndDlgItem, CB_GETCURSEL, 0, 0);
         DC->ConfigNr=(BYTE)SendMessage(hwndDlgItem, CB_GETITEMDATA, dwIndex, 0);

         // Check combination of number of digitizers and selected configuration.
         if (!SeisAD18ConfigOK(DC->NumChannels, DC->ConfigNr)) {
            // User has chosen an invalid configuration.
            wcscpy(Message, TEXT("The selected operating configuration is\n"));
            wcscat(Message, TEXT("not compatible with the selected number\n"));
            wcscat(Message, TEXT("of channels. (See format spesification)\n"));
            MessageBox(hWndDlg, Message, TEXT("Seislog message"), MB_OK|MB_ICONERROR);

            // Move the input focus to the configuration field.
            SetFocus(GetDlgItem(hWndDlg, IDC_COMBO3));

            // Return without closing.
            return TRUE;            
         }

         // Read current choice from the time correction (seconds) edit box.
         GetDlgItemText(hWndDlg, IDC_TIME_CORR, Str20, 20);
         Result=swscanf(Str20, TEXT("%Lf"), &DC->TimeCorrection);
         if (Result==0 || Result==EOF) {
            // A string that cannot be converted to a float value is entered.
            wcscpy(Message, TEXT("The time correction value is missing or invalid.\n"));
            MessageBox(hWndDlg, Message, TEXT("Seislog message"), MB_OK|MB_ICONERROR);

            // Move the input focus to this field.
            SetFocus(GetDlgItem(hWndDlg, IDC_TIME_CORR));
            SendDlgItemMessage(hWndDlg, IDC_TIME_CORR, EM_SETSEL, 0, -1); 

            // Return without closing.
            return TRUE;            
         }

         // Close the dialog box
         EndDialog(hWndDlg , 1);
         return TRUE;
         
         
         case IDCANCEL:  // User has pressed the Cancel button.
         // Do nothing, just exit.
         EndDialog(hWndDlg , 0);
         return TRUE;      
      }
      return FALSE;

   }
   return FALSE;
}


/*******************************************************************/
/* Adjust the time to the nearest whole second.                    */
/*******************************************************************/
VOID Adjust2NearestSec(SYSTEMTIME* Time)
{
   if (Time->wMilliseconds<500) {
      Time->wMilliseconds=0;
   } else {
      Time->wMilliseconds=0;
      AddToTime(Time, 1.0);
   }
}


/*******************************************************************/
/* Configures the device by sending it a number of NULL char's.    */
/*******************************************************************/
VOID ConfigureSeisAD18(HANDLE hComm, BYTE ConfigNr)
{
   BYTE Pulses[34];
   DWORD BytesWritten;

   ZeroMemory((PVOID)Pulses, 34);
   WriteFile(hComm, (PVOID)Pulses, ConfigNr+10, &BytesWritten, NULL); 
}








/***************************************************************************************/
/*                                                                                     */
/* Implementation of the SeisAD18 buffer class starts here.                            */
/*                                                                                     */         
/***************************************************************************************
/***************************************************************************************/
/* Initializes the class.                                                              */
/***************************************************************************************/
VOID SeisAD18Buffer::InitBuffer(BYTE NumChannels, WORD SampleRate) 
{
   SmplRate=SampleRate;       // Save the samplerate.
   NumChnls=NumChannels;      // Save number of channels.
   NumDig=NumChnls/3;         // Calculate number of digitizers.

   // Clear variables and structures.
   CurDigNr=ChnlNr=CurSmplNr=0;
   ZeroMemory((PVOID)&HeaderBytes, sizeof(HeaderBytes));
   ZeroMemory((PVOID)&ChnlData, sizeof(ChnlData));
}


/***************************************************************************************/
/* Clear the buffer contents, but does not clear the info given by call to InitBuffer. */
/***************************************************************************************/
VOID SeisAD18Buffer::ResetBuffer(VOID) 
{
   // Clear variables and structures.
   CurDigNr=ChnlNr=CurSmplNr=0;
   ZeroMemory((PVOID)&HeaderBytes, sizeof(HeaderBytes));
   ZeroMemory((PVOID)&ChnlData, sizeof(ChnlData));
}


/***************************************************************************************/
/* Insert 12 byte block from device into buffer.                                       */
/***************************************************************************************/
VOID SeisAD18Buffer::InsertBuffer(PBYTE Block) 
{
   BYTE *pBytes;
   WORD NumBytes;

   // Insert header bytes into buffer.
   if (HeaderBytes[CurDigNr].NumBytes<MAX_HEADERBYTES) {
      pBytes=HeaderBytes[CurDigNr].Bytes;
      NumBytes=HeaderBytes[CurDigNr].NumBytes;
      CopyMemory((PVOID)(pBytes+NumBytes), (PVOID)Block, 3);
      HeaderBytes[CurDigNr].NumBytes=HeaderBytes[CurDigNr].NumBytes+3;
   }
   
   // Insert sample bytes for first channel.
   ChnlNr=CurDigNr*3;
   pBytes=ChnlData[ChnlNr].Bytes;
   NumBytes=ChnlData[ChnlNr].NumBytes;
   CopyMemory((PVOID)(pBytes+NumBytes), (PVOID)(Block+3), 3);
   ChnlData[ChnlNr].NumBytes=ChnlData[ChnlNr].NumBytes+3;
   HeaderBytes[CurDigNr].ChkSum1=HeaderBytes[CurDigNr].ChkSum1+*(Block+3);
   HeaderBytes[CurDigNr].ChkSum2=HeaderBytes[CurDigNr].ChkSum2+*(Block+4);
   HeaderBytes[CurDigNr].ChkSum3=HeaderBytes[CurDigNr].ChkSum3+*(Block+5);

   // Insert sample bytes for second channel.
   ChnlNr++;
   pBytes=ChnlData[ChnlNr].Bytes;
   NumBytes=ChnlData[ChnlNr].NumBytes;
   CopyMemory((PVOID)(pBytes+NumBytes), (PVOID)(Block+6), 3);
   ChnlData[ChnlNr].NumBytes=ChnlData[ChnlNr].NumBytes+3;
   HeaderBytes[CurDigNr].ChkSum1=HeaderBytes[CurDigNr].ChkSum1+*(Block+6);
   HeaderBytes[CurDigNr].ChkSum2=HeaderBytes[CurDigNr].ChkSum2+*(Block+7);
   HeaderBytes[CurDigNr].ChkSum3=HeaderBytes[CurDigNr].ChkSum3+*(Block+8);

   // Insert sample bytes for third channel.
   ChnlNr++;
   pBytes=ChnlData[ChnlNr].Bytes;
   NumBytes=ChnlData[ChnlNr].NumBytes;
   CopyMemory((PVOID)(pBytes+NumBytes), (PVOID)(Block+9), 3);
   ChnlData[ChnlNr].NumBytes=ChnlData[ChnlNr].NumBytes+3;
   HeaderBytes[CurDigNr].ChkSum1=HeaderBytes[CurDigNr].ChkSum1+*(Block+9);
   HeaderBytes[CurDigNr].ChkSum2=HeaderBytes[CurDigNr].ChkSum2+*(Block+10);
   HeaderBytes[CurDigNr].ChkSum3=HeaderBytes[CurDigNr].ChkSum3+*(Block+11);

   // Advance digitizer nr. Next 12 byte 
   // block comes from the next digitizer.
   CurDigNr++; 
   if (CurDigNr==NumDig) CurDigNr=0;
}


/***************************************************************************************/
/* Returns TRUE if buffer is full (1 second of data for all channels).                 */
/***************************************************************************************/
BOOL SeisAD18Buffer::IsBufferFull(VOID) 
{
   for (Index=0;Index<NumChnls;Index++) {
      if (ChnlData[Index].NumBytes<(SmplRate*3)) return FALSE;
   }
   return TRUE;
}


/***************************************************************************************/
/* Verifyes the given checksums for a given digitizer.                                 */
/***************************************************************************************/
BOOL SeisAD18Buffer::VerifyChkSum(BYTE DigNr, BYTE Chk1, BYTE Chk2, BYTE Chk3)
{
   if (HeaderBytes[DigNr].ChkSum1!=Chk1) return FALSE;
   if (HeaderBytes[DigNr].ChkSum2!=Chk2) return FALSE;
   if (HeaderBytes[DigNr].ChkSum3!=Chk3) return FALSE;
   return TRUE;
}


/***************************************************************************************/
/* Returns the 3 byte-checksums (given in the header) for a given digitizer.           */
/***************************************************************************************/
VOID SeisAD18Buffer::GetChkSum(BYTE DigNr, PBYTE Chk1, PBYTE Chk2, PBYTE Chk3)
{
   DIG_HEADER *Header;
   
   // Extract and return checksums from headerbytes.
   Header=(DIG_HEADER*)HeaderBytes[DigNr].Bytes;
   *Chk1=Header->ChkSum1;
   *Chk2=Header->ChkSum2;
   *Chk3=Header->ChkSum3;
}


/***************************************************************************************/
/* Copies samples (4 bytes pr sample) for one channel to a buffer.                     */
/***************************************************************************************/
VOID SeisAD18Buffer::CopyChnlData(BYTE ChnlNr, PBYTE Buffer)
{
   PBYTE pBytes;
   long int *liPtr;  

   liPtr=(long int*)Buffer;
   for (Index=0;Index<SmplRate;Index++) {
      pBytes=ChnlData[ChnlNr].Bytes+(Index*3);
      *liPtr=*pBytes*65536 + *(pBytes+1)*256 + *(pBytes+2) - 8388608;
      liPtr++;
   }
}


/***************************************************************************************/
/* Returns the GPS status for a digitizer. 0=1PPS signal, 1=1PPS signal missing.       */
/***************************************************************************************/
BYTE SeisAD18Buffer::GetGPSStatus(BYTE DigNr)
{
   DIG_HEADER *Header;
   
   // Extract and return GPS status from headerbytes.
   Header=(DIG_HEADER*)HeaderBytes[DigNr].Bytes;
   return Header->GPSStatus;
}


/***************************************************************************************/
/* Check if the selected digitizer configuration is OK.                                */
/***************************************************************************************/
BOOL SeisAD18ConfigOK(BYTE NumChnls, BYTE ConfigNr)
{
   if (NumChnls==3) return TRUE;
   if (NumChnls==6 || NumChnls==9) {
      if (ConfigNr>=1 && ConfigNr<=6)   return TRUE;
      if (ConfigNr>=10 && ConfigNr<=15) return TRUE;
      if (ConfigNr>=19 && ConfigNr<=24) return TRUE;
   }
   if (NumChnls==12) {
      if (ConfigNr>=4 && ConfigNr<=6)   return TRUE;
      if (ConfigNr>=13 && ConfigNr<=15) return TRUE;
      if (ConfigNr>=19 && ConfigNr<=21) return TRUE;
   }

   return FALSE;
}